use serde::{Deserialize, Serialize};
use serde::de::DeserializeOwned;
use std::collections::HashSet;

pub mod player;
pub mod shop;
pub mod item;
pub mod town;
pub mod farm;
pub mod market;
pub mod token;
pub mod meta;
mod base;
pub mod attraction;

#[derive(Serialize, Deserialize, Debug, Copy, Clone)]
pub enum Dimension {
    Overworld,
    Nether,
    TheEnd
}

impl Default for Dimension {
    fn default() -> Self {
        Self::Overworld
    }
}

#[derive(Serialize, Deserialize, Debug, Copy, Clone)]
pub enum Direction {
    North,
    East,
    South,
    West,
}

#[derive(Default, Serialize, Deserialize, Debug, Copy, Clone)]
pub struct Position {
    pub x: i64,
    pub y: i64,
    pub dimension: Dimension,
}

impl Position {
    pub fn new(x: i64, y: i64, dimension: Dimension) -> Self {
        Self {
            x,
            y,
            dimension
        }
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Tunnel {
    direction: Direction,
    number: i64
}

#[derive(Serialize, Deserialize, Debug, Clone, Default)]
pub struct LocationData {
    name: String,
    position: Position,
    owners: HashSet<u64>,
    tunnel: Option<Tunnel>,
}


impl LocationData {
    pub fn new (name: &str, position: Position, owner: u64, tunnel: Option<Tunnel>) -> Self {
        let mut owners = HashSet::new();
        owners.insert(owner);

        Self {
            name: name.to_string(),
            position,
            owners,
            tunnel,
        }
    }
}

pub trait Location: Serialize + DeserializeOwned + Clone {
    fn new (name: &str, position: Position, owner: u64, tunnel: Option<Tunnel>) -> Self;
    fn mut_location_data(&mut self) -> &mut LocationData;
    fn location_data(&self) -> &LocationData;

    fn name(&self) -> String {
        self.location_data().name.clone()
    }

    fn owners(&self) -> Vec<u64> {
        self.location_data().owners.iter().cloned().collect()
    }

    fn position(&self) -> Position {
        self.location_data().position.clone()
    }

    fn tunnel(&self) -> Option<Tunnel> {
        self.location_data().tunnel.clone()
    }

    fn add_owner(&mut self, owner: u64) {
        self.mut_location_data().owners.insert(owner);
    }

    fn remove_owner(&mut self, owner: u64) {
        self.mut_location_data().owners.remove(&owner);
    }

    fn set_position(&mut self, position: Position) {
        self.mut_location_data().position = position;
    }

    fn set_tunnel(&mut self, tunnel: Tunnel) {
        self.mut_location_data().tunnel = Some(tunnel);
    }

    fn set_name(&mut self, name: &str) {
        self.mut_location_data().name = name.to_string();
    }
}

