use serde::{Deserialize, Serialize};
use chrono::{DateTime, Utc};

use crate::models::item::Item;
use std::collections::HashSet;
use crate::models::{LocationData, Location, Tunnel, Position};
use crate::GeoffreyDatabaseModel;

#[derive(Serialize, Deserialize, Debug, Clone, Hash, Eq, PartialEq)]
pub struct ItemListing {
    pub item: Item,
    pub price: u64,
    pub count_per_price: u64,
    pub restocked_time: DateTime<Utc>
}

impl ItemListing {
    fn new(item: &str, price: u64, count_per_price: u64, restocked_time: DateTime<Utc>) -> Self {
        Self {
            item: Item {name: item.to_string()},
            price,
            count_per_price,
            restocked_time,
        }
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Shop {
    pub id: Option<u64>,
    pub item_listings: HashSet<ItemListing>,
    pub location_data: LocationData
}

impl Location for Shop {
    fn new(name: &str, position: Position, owner: u64, tunnel: Option<Tunnel>) -> Self {
        Self {
            id: None,
            item_listings: Default::default(),
            location_data: LocationData::new(name, position, owner, tunnel)
        }
    }

    fn mut_location_data(&mut self) -> &mut LocationData {
        &mut self.location_data
    }

    fn location_data(&self) -> &LocationData {
        &self.location_data
    }
}

impl GeoffreyDatabaseModel for Shop {
    fn id(&self) -> Option<u64> {
        self.id
    }

    fn set_id(&mut self, id: u64) {
        self.id = Some(id);
    }

    fn tree() -> String {
        "shop".to_string()
    }
}

#[cfg(test)]
mod tests {
    use crate::models::{Position, Dimension, LocationData, Location};
    use crate::models::player::{Player, UserID};
    use crate::models::shop::Shop;

    fn create_shop() -> Shop {
        let player = Player::new("CoolZero123", UserID::DiscordUUID(0));
        let shop = Shop::new("Cool Shop 123", Position::default(), player, None);
    }

    #[test]
    fn test_new_shop() {
    }
}

